{-# OPTIONS_GHC 
         -XRank2Types -XTypeOperators -XEmptyDataDecls
         -XMultiParamTypeClasses -XFlexibleInstances -XFlexibleContexts        
         -XOverlappingInstances
#-}

module Interpreters where

import Prelude hiding (seq)
import AutoLift hiding (put, Put)
import Control.Monad

infixr 5 :+:

data L1 = Nil1 | Put1 Char | Seq1 L1 L1 | Alt1 L1 L1

----------------------------------------------------------------
-- Terms as Free Monads 
----------------------------------------------------------------

data P a = Nil | Put Char | Seq a a | Alt a a 

instance Functor P where
    fmap _ Nil        = Nil
    fmap _ (Put c)    = Put c 
    fmap f (Seq p q)  = Seq  (f p)  (f q)
    fmap f (Alt p q)  = Alt  (f p)  (f q)

data Z a = Num Int | Add a a | Ifz a a a

instance Functor Z where
    fmap f (Num i)      = Num i 
    fmap f (Add p q)    = Add  (f p)  (f q)
    fmap f (Ifz c t e)  = Ifz  (f c)  (f t) (f e)

data E a = Thr | Cat a a

instance Functor E where
    fmap _ Thr        = Thr 
    fmap f (Cat p q)  = Cat  (f p)  (f q)

data Term f x  =  Var x |  Con (f (Term f x))            

instance Functor f  => Functor  (Term f) where
    fmap f (Var x)  = Var (f x)
    fmap f (Con t)  = Con (fmap (fmap f) t)

instance Functor f  => Monad    (Term f) where
    return          = Var
    (Var x) >>= f   = f x
    (Con t) >>= f   = Con (fmap (>>= f) t)  

foldTerm :: (Functor f) => (a -> b) -> (f b -> b) -> Term f a -> b
foldTerm v _ (Var a)    = v a
foldTerm v c (Con fta)  = c (fmap (foldTerm v c) fta)

type Program f = Term f Zero

data Zero

empty :: Zero -> a
empty _ = undefined

----------------------------------------------------------------
-- Coproducts of Free Monads 
----------------------------------------------------------------

data (f :+: g) a     = Inl (f a) | Inr (g a) 

instance (Functor f, Functor g) => Functor (f :+: g) where
    fmap h (Inl fx)  = Inl (fmap h fx)
    fmap h (Inr gx)  = Inr (fmap h gx)

copair                  :: (f a -> b) -> (g a -> b) -> (f :+: g) a -> b
copair  f  _  (Inl fa)  = f fa
copair  _  g  (Inr ga)  = g ga

type L = P :+: Z :+: E
  
prog               :: Program L
prog               = alt1 (seq1 (put1 'a') (add1 (n1 3) (n1 5)))  (cat1 thr1 (put1 'c'))

put1                  :: Char -> Term L a
put1 c                = Con (Inl (Put c))
n1                    :: Int -> Term L a
n1 m                  = Con (Inr (Inl (Num m)))
thr1                  :: Term L a
thr1                  = Con (Inr (Inr Thr))
seq1, alt1,add1,cat1  :: Term L a -> Term L a -> Term L a
seq1 p q              = Con (Inl (Seq p q))
alt1 p q              = Con (Inl (Alt p q))
add1 p q              = Con (Inr (Inl (Add p q)))
cat1 p q              = Con (Inr (Inr (Cat p q)))

type EP   = P :+: E
type EZ   = Z :+: E

throw'  :: Term EP a
throw'  = Con (Inr Thr)

-----------------------------------------------------------------------------
-- SubType relation
-----------------------------------------------------------------------------

infixl 4 >>>>=

class SubType sub sup where
    inj  :: sub a -> sup a
    prj  :: sup a -> Maybe (sub a)
instance SubType f f where
    inj  = id
    prj  = Just

instance SubType f (f :+: g) where
    inj          = Inl
    prj (Inl f)  = Just f
    prj _        = Nothing

instance (SubType f g) => SubType f (h :+: g) where
    inj          = Inr . inj
    prj (Inr a)  = prj a
    prj _        = Nothing

con          :: (SubType s t) => s (Term t x) -> Term t x
con          = Con . inj

nil          :: (SubType P t) => Term t x
nil          = con Nil

put          :: (SubType P s) => Char -> Term s x
put c        = con (Put c)

seq          :: (SubType P s) => Term s x -> Term s x -> Term s x
seq p q      = con (Seq p q)

alt          :: (SubType P s) => Term s x -> Term s x -> Term s x
alt p q      = con (Alt p q)

n            :: (SubType Z s) => Int -> Term s x
n m          = con (Num m)

add          :: (SubType Z s) => Term s x -> Term s x -> Term s x
add p q      = con (Add p q)

ifz          :: (SubType Z s) => Term s x -> (Term s x, Term s x) -> Term s x
ifz c (t,e)  = con (Ifz c t e)

thr          :: (SubType E s) => Term s x
thr          = con Thr

cat          :: (SubType E s) => Term s x -> Term s x -> Term s x
cat p q      = con (Cat p q)

class SubTypeV sub sup where
    injV  :: sub -> sup
    prjV  :: sup -> Maybe sub

instance SubTypeV v v where
    injV  = id
    prjV  = Just

instance SubTypeV v (Either v u) where
    injV           = Left
    prjV (Left v)  = Just v
    prjV _         = Nothing

instance (SubTypeV v w) => SubTypeV v (Either u w) where
    injV            = Right . injV
    prjV (Right a)  = prjV a
    prjV _          = Nothing

(>>>>=)     :: (SubTypeV sub sup, Monad m) => m sup -> (sub -> m sup) -> m sup
m >>>>= f   = do  x <- m 
                  case  prjV x of 
                        Nothing  -> return x
                        Just y   -> f y 

class (Functor f) => InterpC f where
    interp :: f (M V) -> M V

eval :: (InterpC f) => Program f -> M V
eval = foldTerm empty interp

instance (InterpC f, InterpC g) => InterpC (f :+: g) where
    interp = copair interp interp

type V = Either Int ()

type M = ExcT () (WriterT String (NonDetT Id))

instance InterpC Z where
   interp (Num n)      = return (injV n)
   interp (Add x y)    =  x >>>>= \m ->
                          y >>>>= \n -> 
                          return (injV (n + m:: Int))
   interp (Ifz c t e)  = c >>>>= \m -> if (m :: Int)==0 then t else e


instance InterpC E where
   interp Thr        = throw () 
   interp (Cat a h)  = handle a (\() -> h)

instance InterpC P where
   interp Nil        = return (injV ())
   interp (Put c)    = trace [c] >>= return . injV
   interp (Seq t u)  = t >>= \_ ->  u
   interp (Alt t u)  = plusND t u

data R a = Par a a
 
instance Functor R where
    fmap f (Par p q)  = Par  (f p)  (f q)

par          :: (SubType R s) => Term s x -> Term s x -> Term s x
par p q      = con (Par p q)

data Pr a = Pr Char a

instance Functor Pr where
   fmap f (Pr c a) = Pr c (f a)

instance (Monad m) => MonadAlgModel (WriterOp Char) (StepT Pr m) where
    algModel (Trace c a) = step (Pr c (return a))
type M' = ExcT () (StepT Pr NonDet)

instance InterpS R where
   interpS (Par t u)  = g t u
       where g t u =  plusND     (caseStepExcT  (const u) 
                                                (\ft -> step (fmap (`g` u) ft)) t) 
                                 (caseStepExcT  (\v  -> liftM (const v) t) 
                                                (\fu -> step (fmap (g t) fu)) u)
  
caseStepExcT ::  (Functor f, Monad m) =>  
                        (a -> ExcT e (StepT f m) x) -> 
                        (f (ExcT e (StepT f m) a) -> ExcT e (StepT f m) x) -> 
                        ExcT e (StepT f m) a -> ExcT e (StepT f m) x
caseStepExcT v c = X . caseStep (runExcT . either throw v) (runExcT . c . fmap X) . runExcT

-------------------------------------------------------------------------
class Display a where
   display :: a -> String
   display = undefined

instance Display Int where
   display = show

instance Display () where
   display () = "."

-------------------------------------------------------------------------
-- Showing fixpoints
-------------------------------------------------------------------------

class Functor f => PreservesString f where
    preservesString :: f String -> String

instance (Display a, PreservesString f) => Display (Term f a) where
    display (Var a)  = display a
    display (Con x)  = preservesString (fmap display x)

instance (Display a, PreservesString f) => Show (Term f a) where
    show = display

instance (PreservesString f, PreservesString g) => PreservesString (f :+: g) where
    preservesString (Inl x)  = preservesString x
    preservesString (Inr x)  = preservesString x

instance Display Zero where

instance (Display a, Display b) => Display (Either a b) where
   display = either display display

instance PreservesString (Either String) where
   preservesString = either id id

-------------------------------------------------------------------------
-- Displaying Computations
-------------------------------------------------------------------------

instance (Display a, Monad m , PreservesString f, PreservesString m) => Display (StepT f m a) where
   display = preservesString . fmap (either display (preservesString . fmap display)) . runT


instance (Display a, Monad m, PreservesString f, PreservesString m) => Show (StepT f m a) where
   show = display

instance (Monad m, PreservesString f, PreservesString m) => PreservesString (StepT f m) where
   preservesString (T m) = preservesString -- m String
                          (fmap (either id (preservesString. fmap preservesString)) m)


instance PreservesString Pr where
   preservesString (Pr c a) = [c]++a

instance (Display x, Display a, PreservesString m) => Show (ExcT x m a) where
   show = preservesString . fmap display

instance (Display x, PreservesString m) => PreservesString (ExcT x m) where
    preservesString = preservesString . fmap (either (("Exception: "++) . display) id) . runExcT

instance (PreservesString m) => PreservesString (WriterT String m) where
    preservesString = preservesString . fmap (\(s,w) -> w++s) . runWriterT

instance (Monad m, PreservesString m) => PreservesString (NonDetT m) where
    preservesString = preservesString . fmap (("["++) . foldr (\a b -> a++if b=="]" then "]"  else (","++b)) "]")  . collectNonDetT

instance PreservesString Id where
    preservesString = runId

instance Display x => Display (Id x) where
    display = preservesString. fmap display

instance (Display x, PreservesString m) => Display (WriterT String m x) where
    display = preservesString. fmap display

instance (Display x, Display y, PreservesString m) => Display (ExcT x m y) where
    display = preservesString. fmap display

instance (Display x, Monad m, PreservesString m) => Display (NonDetT m x) where
    display = preservesString. fmap display

-------------------------------------------------------------------------
-- Displaying Syntax
-------------------------------------------------------------------------

instance PreservesString Z where
   preservesString (Num n) = display n
   preservesString (Add p q)  = "("++p++" + "++q++")"
   preservesString (Ifz c p q)  = "(if "++c++" then " ++p++" else "++q++")"

instance PreservesString E where
   preservesString Thr = "throw"
   preservesString (Cat p q)  = "(catch "++p++" with "++q++")"

instance PreservesString P where
   preservesString Nil        = "nil"
   preservesString (Put c)    = [c]
   preservesString (Seq p q)  = "("++p++" ; "++q++")"
   preservesString (Alt p q)  = "("++p++" or "++q++")"

instance PreservesString R where
   preservesString (Par p q)  = "("++p++" || "++q++")"


-----------------------------------------------------------
-- Example Programs
-----------------------------------------------------------
type Ex = R :+: L

p0,p1,p2  :: Program L
p0  = seq (put 'b') (put 'c')
p1  = alt (put 'a') (alt p0 (put 'd'))
p2  = seq (alt (put 'a') (put 'd')) p0

p3,p4,p5,p6,p7,p8,p9,p10,p11 :: Program L
p3 = (put 'a' `seq` thr) `alt` (put 'b' `seq` put 'c')
p4 = cat p3 (put 'd')
p5 = (cat (put 'a' `seq` thr) (put 'Y')) `alt` (put 'b' `seq` put 'c')
p6 = cat (put 'z' `seq` p3) (put 'd')
p7 = add p9 p10
p8 = cat (put 'a' `seq` thr) (put 'z' `seq` thr)
p9 = n 1 `add` (n 2 `add` n 4)
p10 = cat (n 1 `add` thr) (n 4 `add` n 6)
p11 = ifz (n 0) (n 1, thr)


p12,p13,p14, p15,p16 :: Program Ex
p12 = (cat (put 'a' `seq` thr) (put 'Y')) `par` (put 'b' `seq` put 'c')
p13 = cat (put 'a' `par` thr) (put 'X')
p14 = (cat (put 'a') (put 'b')) `par` thr
p15 = put 'c' `par` (put 'a' `seq` put 'b') 
p16 = (put 'a' `seq` put 'b') `par` put 'c'

------------------------------
-- Interpreter for M'
------------------------------

class (Functor f) => InterpS f where
    interpS :: f (M' V) -> M' V

evalS :: (InterpS f) => Program f -> M' V
evalS = foldTerm empty interpS

-- Interpreters instances.

instance (InterpS f, InterpS g) => InterpS (f :+: g) where
    interpS = copair interpS interpS

instance InterpS Z where
   interpS (Num n)      = return (injV n)
   interpS (Add x y)    =  x >>>>= \m ->
                           y >>>>= \n -> 
                           return (injV (n + m:: Int))
   interpS (Ifz c t e)  = c >>>>= \m -> if (m :: Int)==0 then t else e

instance InterpS E where
   interpS Thr        = throw () 
   interpS (Cat a h)  = handle a (\() -> h)

instance InterpS P where
   interpS Nil        = return (injV ())
   interpS (Put c)    = trace c >>= return . injV
   interpS (Seq t u)  = t >> u
   interpS (Alt t u)  = plusND t u

